@@ -11,7 +11,7 @@
             [schema.core :as schema]
             [cheshire.core :as cheshire]
             [compojure.core :as compojure :refer [GET ANY PUT]]
-            [liberator.core :as liberator]
+            [liberator.core :refer [defresource]]
             [liberator.representation :as representation]
             [liberator.dev :as liberator-dev]
             [ring.util.response :as rr]))
@@ -112,50 +112,14 @@
         (assoc :status 200)
         (representation/ring-response))))
 
-(defn get-subject
-  "Pull the common name of the subject off the certificate."
-  [certificate]
-  (-> certificate
-      (ca/get-subject)
-      (ca-utils/x500-name->CN)))
-
-(defn client-on-whitelist?
-  "Test if the certificate subject is on the certificate-status client whitelist."
-  [{:keys [access-control]} certificate]
-  (let [whitelist (-> access-control
-                      (get-in [:certificate-status :client-whitelist])
-                      (set))
-        client    (get-subject certificate)]
-    (contains? whitelist client)))
-
-(defn log-rejection
-  "Log a message to info stating that the client is not in the
-   certificate_status access control whitelist."
-  [certificate]
-  (let [subject (get-subject certificate)]
-    (log/info
-      (str "Client '" subject "' access to certificate_status rejected;\n"
-           "client not found in whitelist "
-           "certificate-authority: { certificate-status: { client-whitelist: [...] } }"))))
-
-(defn client-allowed-access?
-  "Determines if the client in the request is allowed to access the
-   certificate_status(es) endpoint based on the client whitelist and
-   whether authorization is required."
-  [settings context]
-  (if (get-in settings [:access-control :certificate-status :authorization-required] true)
-    (if-let [client-cert (get-in context [:request :ssl-client-cert])]
-      (if (client-on-whitelist? settings client-cert)
-        true
-        (do (log-rejection client-cert) false))
-      (log/info "Access to certificate_status rejected; no client certificate found"))
-    true))
-
-(liberator/defresource certificate-status
+(defresource certificate-status
   [subject settings]
   :allowed-methods [:get :put :delete]
 
-  :allowed? (partial client-allowed-access? settings)
+  :allowed? (fn [context]
+              (ringutils/client-allowed-access?
+                (get-in settings [:access-control :certificate-status])
+                (:request context)))
 
   :available-media-types media-types
 
@@ -254,11 +218,14 @@
     (let [desired-state (get-desired-state context)]
       (ca/set-certificate-status! settings subject desired-state))))
 
-(liberator/defresource certificate-statuses
+(defresource certificate-statuses
   [settings]
   :allowed-methods [:get]
 
-  :allowed? (partial client-allowed-access? settings)
+  :allowed? (fn [context]
+              (ringutils/client-allowed-access?
+                (get-in settings [:access-control :certificate-status])
+                (:request context)))
 
   :available-media-types media-types
 
@@ -300,7 +267,7 @@
         (rr/header response "X-Puppet-Version" version)))))
 
 (schema/defn ^:always-validate
-  compojure-app
+  build-ring-handler
   [ca-settings :- ca/CaSettings
    puppet-version :- schema/Str]
   (-> (routes ca-settings)
