@@ -3,12 +3,12 @@
                                      HttpConfiguration HttpConnectionFactory
                                      ConnectionFactory)
            (org.eclipse.jetty.server.handler AbstractHandler ContextHandler HandlerCollection
-                                             ContextHandlerCollection AllowSymLinkAliasChecker)
+                                             ContextHandlerCollection AllowSymLinkAliasChecker StatisticsHandler)
            (org.eclipse.jetty.util.resource Resource)
            (org.eclipse.jetty.util.thread QueuedThreadPool)
            (org.eclipse.jetty.util.ssl SslContextFactory)
            (javax.servlet.http HttpServletRequest HttpServletResponse)
-           (java.util.concurrent Executors)
+           (java.util.concurrent Executors TimeoutException)
            (org.eclipse.jetty.servlets.gzip GzipHandler)
            (org.eclipse.jetty.servlet ServletContextHandler ServletHolder DefaultServlet)
            (org.eclipse.jetty.webapp WebAppContext)
@@ -26,6 +26,7 @@
            (org.eclipse.jetty.util URIUtil))
 
   (:require [ring.util.servlet :as servlet]
+            [ring.util.codec :as codec]
             [clojure.string :as str]
             [clojure.set :as set]
             [clojure.tools.logging :as log]
@@ -60,25 +61,32 @@
    :path schema/Str
    :port schema/Int})
 
-(def ServerIDOption
-  {(schema/optional-key :server-id) schema/Keyword})
+(def CommonOptions
+  {(schema/optional-key :server-id) schema/Keyword
+   (schema/optional-key :redirect-if-no-trailing-slash) schema/Bool})
 
 (def ContextHandlerOptions
-  (assoc ServerIDOption (schema/optional-key :context-listeners) [ServletContextListener]
+  (assoc CommonOptions (schema/optional-key :context-listeners) [ServletContextListener]
                         (schema/optional-key :follow-links) schema/Bool))
 
 (def ServletHandlerOptions
-  (assoc ServerIDOption (schema/optional-key :servlet-init-params) {schema/Str schema/Str}))
+  (assoc CommonOptions (schema/optional-key :servlet-init-params) {schema/Str schema/Str}))
+
+(def ProxySslConfig
+  (merge config/WebserverSslPemConfig
+         {(schema/optional-key :cipher-suites) [schema/Str]
+          (schema/optional-key :protocols)     (schema/maybe [schema/Str])}))
 
 (def ProxyOptions
-  (assoc ServerIDOption
+  (assoc CommonOptions
     (schema/optional-key :scheme) (schema/enum :orig :http :https
                                                "orig" "http" "https")
-    (schema/optional-key :ssl-config) (schema/either
-                                        (schema/eq :use-server-config)
-                                        config/WebserverSslPemConfig)
+    (schema/optional-key :ssl-config) (schema/conditional
+                                        keyword? (schema/eq :use-server-config)
+                                        map?     ProxySslConfig)
     (schema/optional-key :rewrite-uri-callback-fn) (schema/pred ifn?)
     (schema/optional-key :callback-fn) (schema/pred ifn?)
+    (schema/optional-key :failure-callback-fn) (schema/pred ifn?)
     (schema/optional-key :request-buffer-size) schema/Int
     (schema/optional-key :follow-redirects) schema/Bool))
 
@@ -90,28 +98,23 @@
 (def ContextEndpoint
   {:type                                    (schema/eq :context)
    :base-path                               schema/Str
-   (schema/optional-key :context-listeners) (schema/maybe [ServletContextListener])
-   :endpoint                                schema/Str})
+   (schema/optional-key :context-listeners) (schema/maybe [ServletContextListener])})
 
 (def RingEndpoint
-  {:type     (schema/eq :ring)
-   :endpoint schema/Str})
+  {:type     (schema/eq :ring)})
 
 (def ServletEndpoint
   {:type     (schema/eq :servlet)
-   :servlet  java.lang.Class
-   :endpoint schema/Str})
+   :servlet  java.lang.Class})
 
 (def WarEndpoint
   {:type     (schema/eq :war)
-   :war-path schema/Str
-   :endpoint schema/Str})
+   :war-path schema/Str})
 
 (def ProxyEndpoint
   {:type        (schema/eq :proxy)
    :target-host schema/Str
    :target-port schema/Int
-   :endpoint    schema/Str
    :target-path schema/Str})
 
 (def Endpoint
@@ -123,7 +126,12 @@
     #(-> % :type (= :proxy)) ProxyEndpoint))
 
 (def RegisteredEndpoints
-  #{Endpoint})
+  {schema/Str [Endpoint]})
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;;; Constants
+
+(def default-graceful-stop-timeout 60000)
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Utility Functions
@@ -164,12 +172,19 @@
 (schema/defn ^:always-validate
   ssl-context-factory :- SslContextFactory
   "Creates a new SslContextFactory instance from a map of SSL config options."
-  [{:keys [keystore-config client-auth ssl-crl-path ]}
+  [{:keys [keystore-config client-auth ssl-crl-path cipher-suites protocols]}
    :- config/WebserverSslContextFactory]
-  (let [context (SslContextFactory.)]
-    (.setKeyStore context (:keystore keystore-config))
-    (.setKeyStorePassword context (:key-password keystore-config))
-    (.setTrustStore context (:truststore keystore-config))
+  (if (some #(= "sslv3" %) (map str/lower-case protocols))
+    (log/warn (str "`ssl-protocols` contains SSLv3, a protocol with known "
+                  "vulnerabilities; we recommend removing it from the "
+                  "`ssl-protocols` list")))
+
+  (let [context (doto (SslContextFactory.)
+                  (.setKeyStore (:keystore keystore-config))
+                  (.setKeyStorePassword (:key-password keystore-config))
+                  (.setTrustStore (:truststore keystore-config))
+                  (.setIncludeCipherSuites (into-array String cipher-suites))
+                  (.setIncludeProtocols (into-array String protocols)))]
     (if (:trust-password keystore-config)
       (.setTrustStorePassword context (:trust-password keystore-config)))
     (case client-auth
@@ -186,11 +201,13 @@
 
 (schema/defn ^:always-validate
   get-proxy-client-context-factory :- SslContextFactory
-  [ssl-config :- config/WebserverSslPemConfig]
+  [ssl-config :- ProxySslConfig]
   (ssl-context-factory {:keystore-config
                          (config/pem-ssl-config->keystore-ssl-config
                            ssl-config)
-                        :client-auth :none}))
+                        :client-auth :none
+                        :cipher-suites (or (:cipher-suites ssl-config) config/acceptable-ciphers)
+                        :protocols     (or (:protocols ssl-config) config/default-protocols)}))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Jetty Server / Connector Functions
@@ -241,12 +258,11 @@
       (let [ssl-ctxt-factory (ssl-context-factory
                                {:keystore-config (:keystore-config https)
                                 :client-auth     (:client-auth https)
-                                :ssl-crl-path    (:ssl-crl-path https)})
+                                :ssl-crl-path    (:ssl-crl-path https)
+                                :cipher-suites   (:cipher-suites https)
+                                :protocols       (:protocols https)})
             connector        (ssl-connector server ssl-ctxt-factory https)]
-        (when-let [ciphers (:cipher-suites https)]
-          (.setIncludeCipherSuites ssl-ctxt-factory (into-array ciphers)))
-        (when-let [protocols (:protocols https)]
-          (.setIncludeProtocols ssl-ctxt-factory (into-array protocols)))
+
         (.addConnector server connector)
         (swap! (:state webserver-context) assoc :ssl-context-factory ssl-ctxt-factory)))
     server))
@@ -286,7 +302,9 @@
 (schema/defn ^:always-validate
   add-handler :- ContextHandler
   [webserver-context :- ServerContext
-   handler :- ContextHandler]
+   handler :- ContextHandler
+   enable-trailing-slash-redirect?]
+  (.setAllowNullPathInfo handler (not enable-trailing-slash-redirect?))
   (.addHandler (:handlers webserver-context) handler)
   handler)
 
@@ -330,7 +348,7 @@
                                  (:port target)
                                  (with-leading-slash
                                    (URIUtil/addPaths (:path target) context-path))
-                                 query
+                                 (codec/url-decode (str query))
                                  nil)]
             (if-let [rewrite-uri-callback-fn (:rewrite-uri-callback-fn options)]
               (rewrite-uri-callback-fn target-uri req)
@@ -356,13 +374,28 @@
 
       (customizeProxyRequest [proxy-req req]
         (if-let [callback-fn (:callback-fn options)]
-         (callback-fn proxy-req req))))))
+         (callback-fn proxy-req req)))
+
+      (onResponseFailure [req resp proxy-resp failure]
+        (do
+          (log/debug (str (.getRequestId this req) " proxying failed: " failure))
+          (when-not (.isCommitted resp)
+            (if (instance? TimeoutException failure)
+              (.setStatus resp HttpServletResponse/SC_GATEWAY_TIMEOUT)
+              (.setStatus resp HttpServletResponse/SC_BAD_GATEWAY))
+            (when-let [failure-callback-fn (:failure-callback-fn options)]
+              (failure-callback-fn req resp proxy-resp failure))
+            (.complete (.getAttribute req "org.eclipse.jetty.proxy.ProxyServlet.asyncContext")))))
+        )))
 
 (schema/defn ^:always-validate
   register-endpoint!
   [state :- Atom
-   endpoint :- Endpoint]
-  (swap! state update-in [:endpoints] conj endpoint))
+   endpoint-map :- Endpoint
+   endpoint :- schema/Str]
+  (swap! state update-in [:endpoints endpoint] #(if (nil? %)
+                                                   [endpoint-map]
+                                                   (conj % endpoint-map))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Public
@@ -375,7 +408,7 @@
   []
   (let [^ContextHandlerCollection chc (ContextHandlerCollection.)]
     {:handlers chc
-     :state (atom {:endpoints #{}})
+     :state (atom {:endpoints {}})
      :server nil}))
 
 ; TODO move out of public
@@ -442,18 +475,29 @@
    options :- config/WebserverRawConfig]
   {:pre  [(map? options)]
    :post [(started? %)]}
-    (let [config                (config/process-config
-                                  (merge-webserver-overrides-with-options
-                                    webserver-context
-                                    options))
-          ^Server s             (create-server webserver-context config)
-          ^HandlerCollection hc (HandlerCollection.)]
+  (let [config                (config/process-config
+                                (merge-webserver-overrides-with-options
+                                  webserver-context
+                                  options))
+        ^Server s             (create-server webserver-context config)
+        ^HandlerCollection hc (HandlerCollection.)
+        log-handler (config/maybe-init-log-handler options)]
     (.setHandlers hc (into-array Handler [(:handlers webserver-context)]))
-    (let [handler-for-server (if (or (not (contains? options :gzip-enable))
-                                     (:gzip-enable options))
-                               (gzip-handler hc)
-                               hc)]
-      (.setHandler s handler-for-server)
+    (let [shutdown-timeout (when (not (nil? (:shutdown-timeout-seconds options)))
+                             (* 1000 (:shutdown-timeout-seconds options)))
+          maybe-zipped (if (or (not (contains? options :gzip-enable))
+                               (:gzip-enable options))
+                         (gzip-handler hc)
+                         hc)
+          maybe-logged (if log-handler
+                         (doto log-handler (.setHandler hc))
+                         maybe-zipped)
+          statistics-handler (if (or (nil? shutdown-timeout) (pos? shutdown-timeout))
+                               (doto (StatisticsHandler.)
+                                 (.setHandler maybe-logged))
+                               maybe-logged)]
+      (.setHandler s statistics-handler)
+      (.setStopTimeout s (or shutdown-timeout default-graceful-stop-timeout))
       (assoc webserver-context :server s))))
 
 (schema/defn ^:always-validate start-webserver! :- ServerContext
@@ -477,13 +521,16 @@
   add-context-handler :- ContextHandler
   "Add a static content context handler (allow for customization of the context handler through javax.servlet.ServletContextListener implementations)"
   ([webserver-context base-path context-path]
-   (add-context-handler webserver-context base-path context-path nil false))
+   (add-context-handler webserver-context base-path context-path nil {:follow-links? false
+                                                                      :enable-trailing-slash-redirect? false}))
   ([webserver-context :- ServerContext
     base-path :- schema/Str
     context-path :- schema/Str
     context-listeners :- (schema/maybe [ServletContextListener])
-    follow-links?]
-   (let [handler (ServletContextHandler. nil context-path ServletContextHandler/NO_SESSIONS)]
+    options]
+   (let [handler (ServletContextHandler. nil context-path ServletContextHandler/NO_SESSIONS)
+         follow-links? (:follow-links? options)
+         enable-trailing-slash-redirect? (:enable-trailing-slash-redirect? options)]
      (.setBaseResource handler (Resource/newResource base-path))
      (when follow-links?
        (.addAliasCheck handler (AllowSymLinkAliasChecker.)))
@@ -491,16 +538,17 @@
      (when-not (nil? context-listeners)
        (dorun (map #(.addEventListener handler %) context-listeners)))
      (.addServlet handler (ServletHolder. (DefaultServlet.)) "/")
-     (add-handler webserver-context handler))))
+     (add-handler webserver-context handler enable-trailing-slash-redirect?))))
 
 (schema/defn ^:always-validate
   add-ring-handler :- ContextHandler
   [webserver-context :- ServerContext
    handler :- (schema/pred ifn? 'ifn?)
-   path :- schema/Str]
+   path :- schema/Str
+   enable-trailing-slash-redirect?]
   (let [ctxt-handler (doto (ContextHandler. path)
                        (.setHandler (ring-handler handler)))]
-    (add-handler webserver-context ctxt-handler)))
+    (add-handler webserver-context ctxt-handler enable-trailing-slash-redirect?)))
 
 (schema/defn ^:always-validate
   add-servlet-handler :- ContextHandler
@@ -509,13 +557,14 @@
   ([webserver-context :- ServerContext
     servlet :- Servlet
     path :- schema/Str
-    servlet-init-params :- {schema/Any schema/Any}]
+    servlet-init-params :- {schema/Any schema/Any}
+    enable-trailing-slash-redirect?]
    (let [holder   (doto (ServletHolder. servlet)
                     (.setInitParameters servlet-init-params))
          handler  (doto (ServletContextHandler. ServletContextHandler/SESSIONS)
                     (.setContextPath path)
                     (.addServlet holder "/*"))]
-     (add-handler webserver-context handler))))
+     (add-handler webserver-context handler enable-trailing-slash-redirect?))))
 
 (schema/defn ^:always-validate
   add-war-handler :- ContextHandler
@@ -524,11 +573,12 @@
   - `path` is the URL prefix at which the WAR will be registered"
   [webserver-context :- ServerContext
    war :- schema/Str
-   path :- schema/Str]
+   path :- schema/Str
+   disable-redirects-no-slash?]
   (let [handler (doto (WebAppContext.)
                   (.setContextPath path)
                   (.setWar war))]
-    (add-handler webserver-context handler)))
+    (add-handler webserver-context handler disable-redirects-no-slash?)))
 
 (schema/defn ^:always-validate
   add-proxy-route
@@ -540,28 +590,32 @@
   `options` may contain the keys :scheme (legal values are :orig, :http, and
   :https), :ssl-config (value may be :use-server-config or a map containing
   :ssl-ca-cert, :ssl-cert, and :ssl-key), :rewrite-uri-callback-fn (a function
-  taking two arguments, `[target-uri req]`, see README.md/#rewrite-uri-callback-fn)
-  and :callback-fn (a function taking two arguments, `[proxy-req req]`, see
-  README.md/#callback-fn).
+  taking two arguments, `[target-uri req]`, see README.md/#rewrite-uri-callback-fn),
+  :callback-fn (a function taking two arguments, `[proxy-req req]`, see
+  README.md/#callback-fn) and :failure-callback-fn (a function taking four arguments,
+  `[req resp proxy-resp failure]`, see README.md/#failure-callback-fn).
   "
   [webserver-context :- ServerContext
    target :- ProxyTarget
    path :- schema/Str
-   options :- ProxyOptions]
+   options :- ProxyOptions
+   disable-redirects-no-slash?]
   (let [target (update-in target [:path] remove-leading-slash)]
     (add-servlet-handler webserver-context
                          (proxy-servlet webserver-context target options)
-                         path)))
+                         path
+                         {}
+                         disable-redirects-no-slash?)))
 
 (schema/defn ^:always-validate
    get-registered-endpoints :- RegisteredEndpoints
-   "Returns the registered endpoints. Each endpoint is registered upon
-    creation as a map of endpoint information. The type of the
-    endpoint determines what is contained in that endpoint's map of
-    endpoint information. Each map contains the type of the endpoint
-    under the :type key, and the endpoint itself under the :endpoint
-    key.
-
+   "Returns a map of registered endpoints for the given ServerContext.
+    Each endpoint is registered as a key in the map, with its value
+    being an array of maps, each representing a handler registered
+    at that endpoint. Each of these maps contains the type of the
+    handler under the :type key, and may contain additional information
+    as well.
+    
     When the value of :type is :context, the endpoint information will
     be an instance of ContextEndpoint.
 
@@ -709,12 +763,15 @@
         context-listeners (:context-listeners opts)
         s                 (get-server-context context server-id)
         state             (:state s)
-        endpoint          {:type              :context
+        endpoint-map      {:type              :context
                            :base-path         base-path
-                           :context-listeners context-listeners
-                           :endpoint          context-path}]
-    (register-endpoint! state endpoint)
-    (add-context-handler s base-path context-path context-listeners (:follow-links options))))
+                           :context-listeners context-listeners}
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map context-path)
+    (add-context-handler s base-path context-path
+                         context-listeners
+                         {:follow-links?                   (:follow-links options)
+                          :enable-trailing-slash-redirect? enable-redirect})))
 
 (schema/defn ^:always-validate init!
   [context config :- config/WebserverServiceRawConfig]
@@ -746,14 +803,15 @@
       (nil? new-config) (start-server-multiple context config))))
 
 (schema/defn ^:always-validate add-ring-handler!
-  [context handler path options :- ServerIDOption]
-  (let [server-id (:server-id options)
-        s         (get-server-context context server-id)
-        state     (:state s)
-        endpoint  {:type     :ring
-                   :endpoint path}]
-    (register-endpoint! state endpoint)
-    (add-ring-handler s handler path)))
+  [context handler path options :- CommonOptions]
+  (let [server-id     (:server-id options)
+        s             (get-server-context context server-id)
+        state         (:state s)
+        endpoint-map  {:type     :ring}
+
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map path)
+    (add-ring-handler s handler path enable-redirect)))
 
 (schema/defn ^:always-validate add-servlet-handler!
   [context servlet path options :- ServletHandlerOptions]
@@ -763,32 +821,34 @@
         servlet-init-params (:servlet-init-params opts)
         s                   (get-server-context context server-id)
         state               (:state s)
-        endpoint            {:type     :servlet
-                             :servlet  (type servlet)
-                             :endpoint path}]
-    (register-endpoint! state endpoint)
-    (add-servlet-handler s servlet path servlet-init-params)))
+        endpoint-map        {:type     :servlet
+                             :servlet  (type servlet)}
+
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map path)
+    (add-servlet-handler s servlet path servlet-init-params enable-redirect)))
 
 (schema/defn ^:always-validate add-war-handler!
-  [context war path options :- ServerIDOption]
-  (let [server-id (:server-id options)
-        s         (get-server-context context server-id)
-        state     (:state s)
-        endpoint  {:type     :war
-                   :war-path war
-                   :endpoint path}]
-    (register-endpoint! state endpoint)
-    (add-war-handler s war path)))
+  [context war path options :- CommonOptions]
+  (let [server-id     (:server-id options)
+        s             (get-server-context context server-id)
+        state         (:state s)
+        endpoint-map  {:type     :war
+                       :war-path war}
+
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map path)
+    (add-war-handler s war path enable-redirect)))
 
 (schema/defn ^:always-validate add-proxy-route!
   [context target path options]
-  (let [server-id (:server-id options)
-        s         (get-server-context context server-id)
-        state     (:state s)
-        endpoint  {:type        :proxy
-                   :target-host (:host target)
-                   :target-port (:port target)
-                   :target-path (:path target)
-                   :endpoint    path}]
-    (register-endpoint! state endpoint)
-    (add-proxy-route s target path options)))
+  (let [server-id     (:server-id options)
+        s             (get-server-context context server-id)
+        state         (:state s)
+        endpoint-map  {:type        :proxy
+                       :target-host (:host target)
+                       :target-port (:port target)
+                       :target-path (:path target)}
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map path)
+    (add-proxy-route s target path options enable-redirect)))
